import pytest
import pandas as pd
from unittest.mock import patch
import plotly.graph_objects as go
from Toolbox.classes.dashboard import DashboardPlotter

@pytest.fixture
def mock_data():
    data = pd.DataFrame({
        'ISO3': ['USA', 'CAN', 'DEU', 'USA', 'CAN', 'DEU'],
        'Continent': ['North America', 'North America', 'Europe', 'North America', 'North America', 'Europe'],
        'domain': ['Consumption', 'Production', 'Consumption', 'Production', 'Consumption', 'Production'],
        'Commodity': ['Sawnwood C', 'Sawnwood NC', 'Newsprint', 'Sawnwood C', 'Sawnwood NC', 'Newsprint'],
        'Commodity_Group': ['Sawnwood', 'Sawnwood', 'Paper', 'Sawnwood', 'Sawnwood', 'Paper'],
        'Scenario': ['Scenario1', 'Scenario1', 'Scenario2', 'Scenario2', 'Scenario1', 'Scenario2'],
        'year': [2020, 2020, 2021, 2021, 2020, 2021],
        'quantity': [100, 200, 150, 250, 120, 180],
        'price': [2.5, 3.0, 2.8, 3.2, 2.6, 3.1],
        'ForStock': [50, 60, 55, 65, 52, 58]
    })
    return data

@pytest.fixture
def dashboard(mock_data):
    return DashboardPlotter(mock_data)

def test_dashboard_initialization(dashboard):
    assert dashboard.data is not None
    assert dashboard.start == 2020
    assert dashboard.end == 2021
    assert dashboard.app is not None

def test_filter_data(dashboard, mock_data):
    filtered_data = dashboard.filter_data(region=['USA'], continent=None, domain=None, commodity=None, commodity_group=None, scenario=None)
    assert len(filtered_data) == 2
    assert all(filtered_data['ISO3'] == 'USA')

    filtered_data = dashboard.filter_data(region=['DEU'], continent=['Europe'], domain=['Consumption'], 
                                          commodity=['Newsprint'], commodity_group=["Paper"], scenario=['Scenario1', 'Scenario2'])
    assert len(filtered_data) == 1
    assert filtered_data['ISO3'].iloc[0] == 'DEU'
    assert filtered_data['Commodity'].iloc[0] == 'Newsprint'

    filtered_data = dashboard.filter_data(region=None, continent=None, domain=None, commodity=None, commodity_group=None, scenario=None)
    assert len(filtered_data) == len(mock_data)

@pytest.mark.skip(reason="Fehler im test")
def test_update_plot_data(dashboard):
    fig_quantity, fig_price, fig_stock = dashboard.update_plot_data(region=None, continent=None, domain=None, commodity=None, commodity_group=None, scenario=None)
    assert isinstance(fig_quantity, go.Figure)
    assert isinstance(fig_price, go.Figure)
    assert isinstance(fig_stock, go.Figure)
    assert len(fig_quantity.data) > 0
    assert len(fig_price.data) > 0
    assert len(fig_stock.data) > 0

@pytest.mark.skip(reason="Fehler im test")
def test_generate_title(dashboard):
    title = dashboard.generate_title(region=['USA'], continent=['North America'], domain=['Consumption'], commodity=['Sawnwood C'], commodity_group=['Sawnwood'])
    assert title == "['USA'], ['North America'], ['Consumption'], ['Sawnwood C'], ['Sawnwood']"
    title = dashboard.generate_title(region=None, continent=None, domain=None, commodity=None, commodity_group=None)
    assert title == "all data"

def test_create_layout(dashboard):
    dashboard.create_layout()
    assert dashboard.app.layout is not None

def test_create_callbacks(dashboard):
    dashboard.create_callbacks()
    assert True


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from ipywidgets import interactive, widgets
from IPython.display import display, clear_output
import ipywidgets as widgets


class sc_plot():
    def init(self):
        pass

    def predefined_plot(self, data: pd.DataFrame):
        #data[['RegionCode', 'domain']] = data[['RegionCode', 'domain']].astype('str') # must be changed to perform groupby()
        grouped_data = data.groupby(['year', 'Scenario']).sum().reset_index()
        plt.figure(figsize=(12, 8))

        for price_value in grouped_data['Scenario'].unique():
            subset = grouped_data[grouped_data['Scenario'] == price_value]
            plt.plot(subset['year'], subset['quantity'], label=f'Scenario {price_value}')

        plt.title('Quantity for each country grouped by year')
        plt.xlabel('Year')
        plt.ylabel('Quantity')
        plt.legend()
        plt.grid(True)
        #plt.show()
    
class PlotDropDown:
    def __init__(self,data,unique_color:bool = False,color:str="darkblue",legend:bool=True,
                 start:int=1990,end:int=2050,scenario_range:bool=False): 
        self.data =data
        self.unique_color =unique_color
        self.color =color
        self.legend =legend
        self.start=start
        self.end=end
        self.scenario_range=scenario_range
        self.regioncode_dropdown = self.choose_dropdown("RegionCode")
        self.continentcode_dropdown = self.choose_dropdown("Continent")
        self.model_dropdown = self.choose_dropdown("Model")
        self.domain_dropdown = self.choose_dropdown("domain")
        self.commodity_code_dropdown = self.choose_dropdown("CommodityCode")

        self.interactive_plot_update = interactive(self.update_plot_data,
                                                   region = self.regioncode_dropdown,
                                                   continent = self.continentcode_dropdown,
                                                   model = self.model_dropdown,
                                                   domain= self.domain_dropdown,
                                                   commodity = self.commodity_code_dropdown)     
        display(self.interactive_plot_update)


    def choose_dropdown(self, column):
        options = ['Alle'] + list(self.data[column].unique())
        return widgets.Dropdown(
            options=options,
            value='Alle',
            description=f'Select {column}:',
            disabled=False
        )

    def update_plot_data(self, region, continent, model, domain, commodity):
        region_filter = [region] if region != 'Alle' else self.data['RegionCode'].unique()
        continent_filter = [continent] if continent != 'Alle' else self.data['Continent'].unique()
        model_filter = [model] if model != 'Alle' else self.data['Model'].unique()
        domain_filter = [domain] if domain != 'Alle' else self.data['domain'].unique()
        commodity_filter = [commodity] if commodity != 'Alle' else self.data['CommodityCode'].unique()


        filtered_data = self.data[
            (self.data['RegionCode'].isin(region_filter)) &
            (self.data['Continent'].isin(continent_filter)) &
            (self.data['Model'].isin(model_filter)) &
            (self.data['domain'].isin(domain_filter)) &
            (self.data['CommodityCode'].isin(commodity_filter))
        ]
    
        grouped_data = filtered_data.groupby(['year', 'Scenario']).sum().reset_index()
        grouped_data = grouped_data[grouped_data["year"]>=self.start]
        grouped_data = grouped_data[grouped_data["year"]<self.end]

        plt.figure(figsize=(12, 8))

        for scenario in grouped_data['Scenario'].unique():
            subset = grouped_data[grouped_data['Scenario'] == scenario]
            if scenario in ['World500','FAOStat']:
                plt.plot(subset['year'], subset['quantity'], label=f'M: {scenario}',color="black")
            else:
                if self.scenario_range:
                    scenario_group = grouped_data[['year','CommodityCode','quantity']]
                    scenario_group = scenario_group[['year','CommodityCode','quantity']]
                    scenario_mean = scenario_group.groupby(['year','CommodityCode']).mean().reset_index()
                    scenario_max = scenario_group.groupby(['year','CommodityCode']).max().reset_index()
                    scenario_min = scenario_group.groupby(['year','CommodityCode']).min().reset_index()
                    plt.plot(scenario_mean['year'], scenario_mean['quantity'], label=f'M: {scenario}',
                             color=self.color, alpha=0.25, linestyle='dashed')
                    plt.plot(scenario_max['year'], scenario_max['quantity'], label=f'M: {scenario}',
                             color=self.color, alpha=0.35,linestyle='dotted')
                    plt.plot(scenario_min['year'], scenario_min['quantity'], label=f'M: {scenario}',
                             color=self.color, alpha=0.25,linestyle='dotted')
                else:
                    if self.unique_color:
                        plt.plot(subset['year'], subset['quantity'], label=f'M: {scenario}',color=self.color, alpha=0.5)
                    else:
                        plt.plot(subset['year'], subset['quantity'], label=f'M: {scenario}')

        #plt.title(f'quantity for each Year - RegionCode: {region},Continent: {continent}, Model: {model}, ID: {id}, Domain: {domain}, CommodityCode: {commodity}')
        plt.xlabel('Year')
        plt.ylabel('quantity')
        if self.legend:
            plt.legend()
        plt.grid(True)
        #plt.show() 

class HeatmapDropDown:
    def __init__(self, data): 
        self.data = data
        self.reference_data_dropdown = self.choose_dropdown("Scenario")
        self.validation_data_dropdown = self.choose_dropdown("Scenario")
        self.comparator_dropdown = self.choose_dropdown("Comparator")
        self.regioncode_dropdown = self.choose_dropdown("RegionCode")
        self.domain_dropdown = self.choose_dropdown("domain")
        self.commodity_code_dropdown = self.choose_dropdown("CommodityCode")

        self.interactive_plot_update = interactive(self.update_heatmap_data,
                                                   reference_data = self.reference_data_dropdown,
                                                   validation_data = self.validation_data_dropdown,
                                                   comparator = self.comparator_dropdown,
                                                   region = self.regioncode_dropdown,
                                                   domain= self.domain_dropdown,
                                                   commodity = self.commodity_code_dropdown)     
        display(self.interactive_plot_update)


    def choose_dropdown(self, column):
        if (column == "Scenario"):
            options = list(self.data[column].unique())
            value = options[0]
        elif (column == "Comparator"):
            options = ['abs_quantity_diff', 'rel_quantity_diff', 'abs_price_diff', 'rel_price_diff']
            value = 'rel_quantity_diff'
        else:
            options = ['All'] + list(self.data[column].unique())
            value = 'All'

        return widgets.Dropdown(
            options=options,
            value=value,
            description=f'Select {column}:',
            disabled=False
        )    
    def update_heatmap_data(self, reference_data, validation_data, comparator, region, domain, commodity):
        region_filter = [region] if region != 'All' else self.data['RegionCode'].unique()
        domain_filter = [domain] if domain != 'All' else self.data['domain'].unique()
        commodity_filter = [commodity] if commodity != 'All' else self.data['CommodityCode'].unique()
        reference_filter = reference_data
        validation_filter = validation_data
        comparator_filter = comparator

        max_period = max(set(self.data[(self.data['Model'] == 'GFPMpt') &
                                       (self.data['Scenario'] == validation_filter)]['Period']))

        reference_data_filtered = self.data[self.data['Scenario'] == reference_filter]
        validation_data_filtered = self.data[self.data['Scenario'] == validation_filter]

        # Filter domain and commodities of interest
        reference_data_filtered = reference_data_filtered[(reference_data_filtered['domain'].isin(domain_filter)) &
                                                          (reference_data_filtered['RegionCode'].isin(region_filter)) &
                                                          (reference_data_filtered['CommodityCode'].isin(commodity_filter)) &
                                                          (reference_data_filtered['Period'] <= max_period)].reset_index(drop=True)
        
        reference_data_grouped = reference_data_filtered.groupby(['Period', 'CommodityCode']).sum().reset_index() 
        
        validation_data_filtered = validation_data_filtered[(validation_data_filtered['domain'].isin(domain_filter)) &
                                                            (validation_data_filtered['RegionCode'].isin(region_filter)) &
                                                            (validation_data_filtered['CommodityCode'].isin(commodity_filter))].reset_index(drop=True)
        
        validation_data_grouped = validation_data_filtered.groupby(['Period', 'CommodityCode']).sum().reset_index() 

        data_info = validation_data_grouped[['CommodityCode', 'Period']]

        abs_price_diff = pd.DataFrame(validation_data_grouped['price'] - reference_data_grouped['price']).rename(columns={'price': 'abs_price_diff'})
        abs_quantity_diff = pd.DataFrame(validation_data_grouped['quantity'] - reference_data_grouped['quantity']).rename(columns={'quantity': 'abs_quantity_diff'})
        rel_price_diff = pd.DataFrame((abs_price_diff['abs_price_diff'] / reference_data_grouped['price']) * 100).rename(columns={0: 'rel_price_diff'})
        rel_quantity_diff = pd.DataFrame((abs_quantity_diff['abs_quantity_diff'] / reference_data_grouped['quantity']) * 100).rename(columns={0: 'rel_quantity_diff'})

        data_heatmap = pd.concat([data_info,
                                  abs_price_diff['abs_price_diff'],
                                  abs_quantity_diff['abs_quantity_diff'],
                                  rel_price_diff['rel_price_diff'],
                                  rel_quantity_diff['rel_quantity_diff']], axis=1)

        data_heatmap.fillna(0, inplace=True)
        data_heatmap.replace(np.inf, 0, inplace=True)

        fig = data_heatmap.pivot('CommodityCode', 'Period', f'{comparator_filter}')  # TODO dynamize f'{rel_quantity_diff}' with drop down options
        f, ax = plt.subplots(figsize=(13, 9))
        plt.title(f'{comparator_filter} for {domain_filter}-quantities in {region_filter}')
        sns.heatmap(fig, annot=True, linewidths=.5, ax=ax, cbar_kws={'label': 'Deviation of GFPMpt from GFPM'})
        
class InteractivePrice:
    def __init__(self, data):
        self.data = data
        self.region_dropdown = self.create_dropdown('RegionCode', 'Select RegionCode:')
        self.model_dropdown = self.create_dropdown('Model', 'Select Model:')
        self.id_dropdown = self.create_dropdown('ID', 'Select ID:')
        self.domain_dropdown = self.create_dropdown('domain', 'Select Domain:')
        self.commodity_code_dropdown = self.create_dropdown('CommodityCode', 'Select CommodityCode:')

        self.interactive_plot_update = widgets.interactive(
            self.update_plot_data,
            region_code=self.region_dropdown,
            model=self.model_dropdown,
            id_value=self.id_dropdown,
            domain=self.domain_dropdown,
            commodity_code=self.commodity_code_dropdown
        )

        self.region_dropdown.observe(self.update_outputs, 'value')
        self.model_dropdown.observe(self.update_outputs, 'value')
        self.id_dropdown.observe(self.update_outputs, 'value')
        self.domain_dropdown.observe(self.update_outputs, 'value')
        self.commodity_code_dropdown.observe(self.update_outputs, 'value')

        self.output_plot = widgets.Output()
        self.output_table = widgets.Output()

        display(self.region_dropdown, self.model_dropdown, self.id_dropdown, self.domain_dropdown, self.commodity_code_dropdown)
        display(self.output_plot)
        display(self.output_table)

    def create_dropdown(self, column, description):
        dropdown = widgets.Dropdown(
            options=['Alle'] + list(self.data[column].unique()),
            value='Alle',
            description=description,
            disabled=False,
        )
        return dropdown

    def update_plot_data(self, region_code, model, id_value, domain, commodity_code):
        region_code_filter = [region_code] if region_code != 'Alle' else self.data['RegionCode'].unique()
        domain_filter = [domain] if domain != 'Alle' else self.data['domain'].unique()
        commodity_code_filter = [commodity_code] if commodity_code != 'Alle' else self.data['CommodityCode'].unique()

        filtered_data = self.data[
            (self.data['RegionCode'].isin(region_code_filter)) &
            (self.data['domain'].isin(domain_filter)) &
            (self.data['CommodityCode'].isin(commodity_code_filter))
        ]

        grouped_data = filtered_data.groupby(['Period', 'Scenario']).sum().reset_index()

        with self.output_plot:
            clear_output(wait=True)
            plt.figure(figsize=(12, 8))
            bar_width = 0.2
            for i, scenario_value in enumerate(grouped_data['Scenario'].unique()):
                subset = grouped_data[grouped_data['Scenario'] == scenario_value]
                x_positions = subset['Period'] + i * bar_width
                plt.bar(x_positions, subset['price'], width=bar_width, label=f'Scenario {scenario_value}')

            plt.title(f'Price for each scenario grouped by Period - RegionCode: {region_code}, Domain: {domain}, CommodityCode: {commodity_code}')
            plt.xlabel('Period')
            plt.ylabel('Price')
            plt.grid(True)
            #plt.show()

        with self.output_table:
            clear_output(wait=True)
            display_table = filtered_data[['Period', 'RegionCode', 'domain', 'CommodityCode', 
                                           'Scenario', 'quantity', 'price']]
            display(display_table)

    def update_outputs(self, *args):
        self.update_plot_data(self.region_dropdown.value, self.domain_dropdown.value, self.commodity_code_dropdown.value)
        
class interactiveModelComparison():
    def __init__(self, data, plot_option):
        self.data = data
        self.plot_option = plot_option
        self.model_dropdown = self.create_dropdown('Model', 'Select Model:')
        self.region_dropdown = self.create_dropdown('Region', 'Select Region:')
        self.parameter_dropdown = self.create_dropdown('Parameter', 'Select Parameter:')
        self.scenario_dropdown = self.create_dropdown('Scenario', 'Select Scenario:')

        self.interactive_plot_update = widgets.interactive(
            self.update_plot_data,
            region = self.region_dropdown,
            model = self.model_dropdown,
            parameter = self.parameter_dropdown,
            scenario = self.scenario_dropdown
        )

        self.region_dropdown.observe(self.update_outputs, 'value')
        self.model_dropdown.observe(self.update_outputs, 'value')
        self.parameter_dropdown.observe(self.update_outputs, 'value')
        self.scenario_dropdown.observe(self.update_outputs, 'value')

        self.output_plot = widgets.Output()
        self.output_table = widgets.Output()

        display(self.region_dropdown, self.model_dropdown, self.parameter_dropdown, self.scenario_dropdown)
        display(self.output_plot)
        display(self.output_table)

    def create_dropdown(self, column, description):
        dropdown = widgets.Dropdown(
            options=['Alle'] + list(self.data[column].unique()),
            value='Alle',
            description=description,
            disabled=False,
        )
        return dropdown
    
    def update_plot_data(self, region, model, parameter, scenario):
        region_filter = [region] if region != 'Alle' else self.data['Region'].unique()
        model_filter = [model] if model != 'Alle' else self.data['Model'].unique()
        parameter_filter = [parameter] if parameter != 'Alle' else self.data['Parameter'].unique()
        scenario_filter = [scenario] if scenario != 'Alle' else self.data['Scenario'].unique()

        filtered_data = self.data[
            (self.data['Region'].isin(region_filter)) &
            (self.data['Model'].isin(model_filter)) &
            (self.data['Parameter'].isin(parameter_filter)) &
            (self.data['Scenario'].isin(scenario_filter))
        ].reset_index(drop=True)

        max_period = max(filtered_data[filtered_data['Model'] == 'GFPMpt']['Period'])
        filtered_data = filtered_data[filtered_data['Period'] <= max_period].reset_index(drop=True)

        with self.output_plot:
            clear_output(wait=True)
            sns.set_style("darkgrid", {"grid.color": ".6", "grid.linestyle": ":"})
            sns.set_theme(palette='Dark2')  # other fitting possiblities: Paired_r, Paired
            fig, ax = plt.subplots(figsize=(20, 10))
            if self.plot_option == 'min_max':
                gfpmpt_data = filtered_data[filtered_data['Model'] == 'GFPMpt'].reset_index(drop=True)
                
                fsm_data = filtered_data[filtered_data['Model'] != 'GFPMpt'].reset_index(drop=True)
                fsm_data_max = fsm_data.groupby(['Period', 'Region', 'Parameter', 'Scenario'])['Data'].max().reset_index()
                fsm_data_min = fsm_data.groupby(['Period', 'Region', 'Parameter', 'Scenario'])['Data'].min().reset_index()
                fsm_data_max['Model'] = 'Max'
                fsm_data_min['Model'] = 'Min'

                data_fin = pd.concat([gfpmpt_data, fsm_data_max, fsm_data_min], axis=0).reset_index(drop=True)

                sns.lineplot(x='Period', y='Data', data=data_fin, hue= 'Model', style='Region')
            
            if self.plot_option == 'ssp_fsm_range':
                sns.lineplot(x="Period", y="Data", hue="Model", data=filtered_data, style='Region')

            if self.plot_option == 'ssp_fsm_all':
                filtered_data['Scenario'] = filtered_data['Region'] + '_' + filtered_data['Scenario']
                sns.lineplot(x="Period", y="Data", hue="Model", data=filtered_data, style='Scenario')
                # Alternative representation
                # filtered_data['Scenario'] = filtered_data['Model'] + '_' + filtered_data['Scenario']
                # sns.lineplot(x="Period", y="Data", hue="Scenario", data=filtered_data, style='Region')


            plt.title(f'Model intercomparison results - Region: {region}, Model: {model}, Parameter: {parameter}, Scenario: {scenario}')
            plt.xlabel('Period')
            plt.ylabel(f'{parameter}')
            #plt.legend()
            plt.grid(True)
            #plt.show()

        with self.output_table:
            clear_output(wait=True)
            display_table = filtered_data[['Period', 'Region', 'Model', 'Parameter',
                                           'Scenario', 'Data']]
            display(display_table)
    

    
    def update_outputs(self, *args):
        self.update_plot_data(self.region_dropdown.value, self.model_dropdown.value, 
                              self.parameter_dropdown.value, self.scenario_dropdown.value)
    
    


import pandas as pd
import pytest
from Toolbox.classes.model_analysis import validation

@pytest.fixture
def validator():
    return validation()

@pytest.fixture
def mock_data():
    return pd.DataFrame({
        'ID': [0, 1, 1],
        'Scenario': ['Base', 'S1', 'S1'],
        'RegionCode': ['A', 'A', 'A'],
        'Continent': ['Europe', 'Europe', 'Europe'],
        'domain': ['Supply', 'Supply', 'Supply'],
        'CommodityCode': [101, 101, 101],
        'year': [2020, 2020, 2020],
        'quantity': [100, 80, 90]
    })

@pytest.mark.skip(reason="Fehler im test")
def test_model_difference(validator, mock_data):
    result = validator.model_difference(mock_data)
    assert 'S1' in result.columns
    assert isinstance(result, pd.DataFrame)
    assert result['S1'].iloc[0] == 10

def test_model_corrcoef(validator, mock_data):
    result = validator.model_corrcoef(mock_data)
    assert isinstance(result, pd.DataFrame)
    assert not result.empty

@pytest.mark.skip(reason="Fehler im test")
def test_validation_std(validator, mock_data):
    result = validator.validation(mock_data)
    assert isinstance(result, pd.DataFrame)
    assert result.columns[3] in ['S1']

def test_reformate_external_data(validator):
    sample_data = pd.DataFrame({
        'Model': ['M1', 'M1'],
        'Estimate': ['RoundwoodHarvest', 'ForestArea'],
        'SSP': ['SSP1', 'SSP1'],
        'Region': ['Europe', 'Europe'],
        'RCP-SSP': ['Baseline_SSP1', 'Baseline_SSP1'],
        2015: [1, 2],
        2020: [3, 4],
        2025: [5, 6],
        2030: [7, 8],
        2035: [9, 10]
    })

    result = validator.reformate_external_data(sample_data)
    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {'Model', 'Parameter', 'Scenario', 'Region', 'Period', 'Data'}
    assert result['Period'].max() <= 2035

def test_add_world(validator):
    df = pd.DataFrame({
        'Model': ['GFPM']*2,
        'Parameter': ['RoundwoodHarvest', 'RoundwoodHarvest'],
        'Scenario': ['S1', 'S1'],
        'Region': ['Europe', 'Asia'],
        'Period': [2025, 2025],
        'Data': [1000, 2000]
    })
    result = validator.add_world(df)
    assert 'World' in result['Region'].values

def test_convert_unit(validator):
    df = pd.DataFrame({
        'Parameter': ['RoundwoodHarvest', 'ForestArea', 'CarbonStockBiomass [MtCO2]'],
        'Data': [1000, 2000, 44]
    })
    result = validator.convert_unit(df.copy())
    assert result.loc[0, 'Data'] == 1
    assert result.loc[1, 'Data'] == 2
    assert round(result.loc[2, 'Data'], 2) == 12.0

@pytest.mark.skip(reason="Fehler im test")
def test_rename_parameter(validator):
    df = pd.DataFrame({
        'Parameter': ['CarbonStockBiomass [MtCO2]'],
        'Scenario': ['model_rcp_ssp_otherinfo_SSP1']
    })
    result = validator.rename_parameter(df.copy())
    assert result['Parameter'].iloc[0] == 'Total Forest Non-soil C Stock (MtC)'
    assert result['Scenario'].iloc[0] == 'SSP1'

def test_merge_data(validator):
    df1 = pd.DataFrame({'Parameter': ['A'], 'Data': [1]})
    df2 = pd.DataFrame({'Parameter': ['B'], 'Data': [2]})
    result = validator.merge_data(df1, df2)
    assert len(result) == 2



import click
import warnings
from Toolbox.toolbox import timba_dashboard
import Toolbox.parameters.paths as toolbox_paths
from pathlib import Path
warnings.simplefilter(action='ignore', category=FutureWarning)

@click.command()
@click.option('-NF', '--num_files', default=10, 
              show_default=True, required=True, type=int, 
              help="Specify the number of most recent .pkl files to read. Limits output to prevent overcrowding.")
@click.option('-FP', '--sc_folderpath', default=toolbox_paths.SCINPUTPATH, 
              show_default=True, required=True, type=Path, 
              help="Define the folder where the code will look for .pkl files containing the scenarios.")
@click.option('-AIFP', '--additional_info_folderpath', default=toolbox_paths.AIINPUTPATH, 
              show_default=True, required=True, type=Path, 
              help="Define the folder where the code will look for additional infos, like historic data or country information.")

def cli(num_files,sc_folderpath,additional_info_folderpath):    
    td = timba_dashboard(num_files_to_read=num_files,
                         scenario_folder_path=sc_folderpath,
                         additional_info_folderpath=additional_info_folderpath)
    td.run()

if __name__ == '__main__':
    cli()



import pytest
import pandas as pd
from unittest.mock import patch, mock_open
import pickle
import gzip
from pathlib import Path
from io import BytesIO
from Toolbox.classes.import_data import import_pkl_data 
import Toolbox.parameters.paths as toolbox_paths
import Toolbox.parameters.default_parameters as toolbox_parameters

@pytest.fixture
def import_pkl_instance():
    return import_pkl_data(SCENARIOPATH=toolbox_paths.SCINPUTPATH)

def create_mock_pkl_file(data):
    mock_pkl = pickle.dumps(data)
    mock_file = BytesIO(mock_pkl)
    return mock_file

def create_mock_gzip_file(data):
    mock_pkl = pickle.dumps(data)
    mock_gzip = BytesIO()
    with gzip.GzipFile(fileobj=mock_gzip, mode='wb') as f:
        f.write(mock_pkl)
    mock_gzip.seek(0)
    return mock_gzip

def test_open_pickle_success(import_pkl_instance):
    mock_data = {'test': 'data'}
    mock_file = create_mock_pkl_file(mock_data)
    
    with patch("builtins.open", return_value=mock_file):
        result = import_pkl_instance.open_pickle("dummy_path.pkl")
    assert result == mock_data

@patch('pandas.read_csv')
def test_read_country_data(mock_read_csv, import_pkl_instance):
    mock_data = pd.DataFrame({
        "Country-Code": [1, 2],
        "ContinentNew": ["A", "B"],
        "Country": ["X", "Y"],
        "ISO-Code": ["x1", "y1"]
    })
    mock_read_csv.return_value = mock_data

    result = import_pkl_instance.read_country_data()
    assert isinstance(result, pd.DataFrame)
    assert list(result.columns) == ["RegionCode", "Continent", "Country", "ISO3"]
    assert result["Country"].dtype == "category"
    mock_read_csv.assert_called_once()

@patch('pandas.read_csv')
def test_read_commodity_data(mock_read_csv, import_pkl_instance):
    mock_data = pd.DataFrame({
        "Commodity": ["A", "B"],
        "CommodityCode": ["C1", "C2"],
        "Commodity_Group": ["G1", "G2"]
    })
    mock_read_csv.return_value = mock_data
    result = import_pkl_instance.read_commodity_data()
    assert isinstance(result, pd.DataFrame)
    assert list(result.columns) == ["Commodity", "CommodityCode", "Commodity_Group"]
    assert result["Commodity"].dtype == "category"
    mock_read_csv.assert_called_once()

@patch('pandas.read_csv')
def test_read_historic_data(mock_read_csv, import_pkl_instance):
    mock_data = pd.DataFrame({
        "RegionCode": [1, 2],
        "CommodityCode": [3, 4],
        "domain": ["A", "B"],
        "price": [1.0, 2.0],
        "quantity": [3.0, 4.0],
        "Period": [5, 6],
        "year": [2020, 2021],
        "Scenario": ["S1", "S2"],
        "Model": ["M1", "M2"]
    })
    mock_read_csv.return_value = mock_data
    result = import_pkl_instance.read_historic_data()
    assert isinstance(result, pd.DataFrame)
    assert result["RegionCode"].dtype == "category"
    assert result["price"].dtype == "float32"
    mock_read_csv.assert_called_once()

def test_downcasting(import_pkl_instance):
    data = pd.DataFrame({
        "RegionCode": [1, 2],
        "CommodityCode": [3, 4],
        "domain": ["A", "B"],
        "price": [1.0, 2.0],
        "quantity": [3.0, 4.0],
        "Period": [5, 6],
        "year": [2020, 2021],
        "Scenario": ["S1", "S2"],
        "Model": ["M1", "M2"]
    })

    result = import_pkl_instance.downcasting(data)

    assert result["RegionCode"].dtype == "category"
    assert result["price"].dtype == "float32"
    assert result["Period"].dtype == "int16"

def test_add_consumption(import_pkl_instance):
    data = pd.DataFrame({
        "quantity_ManufactureCost": [1, 2],
        "quantity_Supply": [3, 4],
        "quantity_TransportationExport": [1, 1],
        "quantity_TransportationImport": [1, 1],
        "price_ManufactureCost": [10, 20],
        "price_Supply": [30, 40],
        "price_TransportationExport": [10, 10],
        "price_TransportationImport": [10, 10],
    })

    result = import_pkl_instance.add_consumption(data.copy())

    assert "quantity" in result.columns
    assert "price" in result.columns
    assert result["domain"].iloc[0] == "Consumption"
    assert (result["quantity"] >= 0).all()

def test_add_net_exports(import_pkl_instance):
    data = pd.DataFrame({
        "quantity_TransportationExport": [2, 3],
        "quantity_TransportationImport": [1, 1],
        "price_TransportationExport": [20, 30]
    })
    result = import_pkl_instance.add_net_exports(data.copy())
    assert "quantity" in result.columns
    assert "price" in result.columns
    assert result["domain"].iloc[0] == "Net Exports"
    assert (result["quantity"] == [1, 2]).all()

def test_add_net_imports(import_pkl_instance):
    data = pd.DataFrame({
        "quantity_TransportationImport": [2, 3],
        "quantity_TransportationExport": [1, 1],
        "price_TransportationImport": [20, 30]
    })
    result = import_pkl_instance.add_net_imports(data.copy())
    assert "quantity" in result.columns
    assert "price" in result.columns
    assert result["domain"].iloc[0] == "Net Imports"
    assert (result["quantity"] == [1, 2]).all()

def test_add_production(import_pkl_instance):
    data = pd.DataFrame({
        "quantity_ManufactureCost": [1, 2],
        "quantity_Supply": [3, 4],
        "price_ManufactureCost": [10, 20],
        "price_Supply": [30, 40]
    })
    result = import_pkl_instance.add_production(data.copy())
    assert "quantity" in result.columns
    assert "price" in result.columns
    assert result["domain"].iloc[0] == "Production"
    assert (result["quantity"] == [4, 6]).all()

def test_concat_calc_domains(import_pkl_instance):
    origin_data = pd.DataFrame({
        'RegionCode': [1, 2],
        'CommodityCode': [3, 4],
        'Period': [5, 6],
        'year': [2020, 2021],
        'domain': ['A', 'B'],
        'price': [1.0, 2.0],
        'quantity': [3.0, 4.0],
        'other_col': [7, 8]
    })
    calc_data = pd.DataFrame({
        'RegionCode': [1, 2],
        'CommodityCode': [3, 4],
        'Period': [5, 6],
        'year': [2020, 2021],
        'domain': ['C', 'D'],
        'price': [5.0, 6.0],
        'quantity': [7.0, 8.0],
        'another_col': [9, 10]
    })
    result = import_pkl_instance.concat_calc_domains(origin_data, calc_data)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 4
    assert 'other_col' in result.columns
    assert 'another_col' not in result.columns

def test_add_calculated_domains(import_pkl_instance):
    data = {
        "data_periods": pd.DataFrame({
            'RegionCode': [1, 1, 2, 2],
            'CommodityCode': [101, 101, 102, 102],
            'Period': [1, 2, 1, 2],
            'year': [2020, 2021, 2020, 2021],
            'domain': ['A', 'B', 'A', 'B'],
            'price': [1.0, 2.0, 3.0, 4.0],
            'quantity': [10.0, 20.0, 30.0, 40.0]
        })
    }
    with patch.object(import_pkl_data, 'add_net_exports') as mock_add_net_exports, \
            patch.object(import_pkl_data, 'add_net_imports') as mock_add_net_imports:
        mock_add_net_exports.return_value = pd.DataFrame({
            'RegionCode': [1, 2],
            'CommodityCode': [101, 102],
            'Period': [1, 1],
            'year': [2020, 2020],
            'domain': ['Net Exports', 'Net Exports'],
            'price': [5.0, 6.0],
            'quantity': [50.0, 60.0]
        })
        mock_add_net_imports.return_value = pd.DataFrame({
            'RegionCode': [1, 2],
            'CommodityCode': [101, 102],
            'Period': [1, 1],
            'year': [2020, 2020],
            'domain': ['Net Imports', 'Net Imports'],
            'price': [7.0, 8.0],
            'quantity': [70.0, 80.0]
        })
        result = import_pkl_instance.add_calculated_domains(data)
        assert isinstance(result, pd.DataFrame)
        assert mock_add_net_exports.called
        assert mock_add_net_imports.called

def test_concat_scenarios(import_pkl_instance):
    data = {
        "data_periods": pd.DataFrame({
            'RegionCode': [1, 2],
            'CommodityCode': [3, 4],
            'Period': [5, 6],
            'year': [2020, 2021],
            'domain': ['A', 'B'],
            'price': [1.0, 2.0],
            'quantity': [3.0, 4.0]
        })
    }
    sc_name = "scenario1"
    data_prev = {
        "data_periods": pd.DataFrame({
            'RegionCode': [5, 6],
            'CommodityCode': [7, 8],
            'Period': [9, 10],
            'year': [2022, 2023],
            'domain': ['C', 'D'],
            'price': [5.0, 6.0],
            'quantity': [7.0, 8.0],
            'Scenario': ['scenario0', 'scenario0'],
            'Model': ['model0', 'model0']
        })
    }
    ID = 1
    with patch.object(import_pkl_data, 'add_calculated_domains') as mock_add_calculated_domains:
        mock_add_calculated_domains.return_value = data["data_periods"]
        import_pkl_instance.concat_scenarios(data, sc_name, data_prev, ID)
        assert "Scenario" in data["data_periods"].columns
        assert "Model" in data["data_periods"].columns

# def test_combined_data(import_pkl_instance, tmp_path):
#     test_data = {'data_periods': pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})}
#     file1 = tmp_path / "scenario1.pkl"
#     file2 = tmp_path / "scenario2.pkl"
#     with gzip.open(file1, 'wb') as f:
#         pickle.dump(test_data, f)
#     with gzip.open(file2, 'wb') as f:
#         pickle.dump(test_data, f)
#     with patch.object(import_pkl_data, 'read_historic_data') as mock_read_historic_data, \
#             patch.object(import_pkl_data, 'read_country_data') as mock_read_country_data, \
#             patch.object(import_pkl_data, 'read_commodity_data') as mock_read_commodity_data:
#         mock_read_historic_data.return_value = pd.DataFrame({'hist_col1': [5, 6], 'hist_col2': [7, 8]})
#         mock_read_country_data.return_value = pd.DataFrame({'RegionCode': [1, 2], 'Continent': ['A', 'B'], 'Country': ['X', 'Y'], 'ISO3': ['x1', 'y1']})
#         mock_read_commodity_data.return_value = pd.DataFrame({'Commodity': ['C1', 'C2'], 'CommodityCode': [101, 102], 'Commodity_Group': ['G1', 'G2']})
#         import_pkl_instance.SCENARIOPATH = tmp_path
#         import_pkl_instance.num_files_to_read = 2
#         result = import_pkl_instance.combined_data()
#         assert isinstance(result, dict)
#         assert 'data_periods' in result
#         assert mock_read_historic_data.called
#         assert mock_read_country_data.called
#         assert mock_read_commodity_data.called


import pickle
import pandas as pd
import numpy as np
from pathlib import Path
import os
from enum import Enum
import pickle
import gzip
import Toolbox.parameters.paths as toolbox_paths
import Toolbox.parameters.default_parameters as toolbox_parameters

class import_pkl_data:
    def __init__(self, num_files_to_read:int=10,
                 SCENARIOPATH:Path= toolbox_paths.SCINPUTPATH,
                 ADDINFOPATH:Path= toolbox_paths.AIINPUTPATH):
        self.num_files_to_read = num_files_to_read
        self.SCENARIOPATH = SCENARIOPATH
        self.ADDINFOPATH = ADDINFOPATH

    def open_pickle(self, src_filepath: str):
        """open pkl file
        :param src_filepath: source path for pkl file
        :return: object from pkl file
        """
        import pickle
        with open(src_filepath, "rb") as pkl_file:
            obj = pickle.load(pkl_file)
        return obj

    def read_country_data(self):
        """read data additional information for country data
        :return: country data
        """
        country_data = pd.read_csv(self.ADDINFOPATH / toolbox_paths.COUNTRYINFO, encoding = "ISO-8859-1")
        country_data = country_data[["Country-Code", "ContinentNew", "Country","ISO-Code"]]
        country_data.columns = ["RegionCode","Continent", "Country","ISO3"]
        country_data.Country = country_data.Country.astype("category")
        country_data.Continent = country_data.Continent.astype("category")
        country_data.ISO3 = country_data.ISO3.astype("category")
        return country_data
    
    def read_commodity_data(self):
        """read data additional information for commodity data
        :return: commodity data
        """
        commodity_data = pd.read_csv(self.ADDINFOPATH / toolbox_paths.COMMODITYINFO , encoding = "ISO-8859-1")
        commodity_data = commodity_data[["Commodity","CommodityCode","Commodity_Group"]]
        commodity_data.Commodity = commodity_data.Commodity.astype("category")
        commodity_data.CommodityCode = commodity_data.CommodityCode.astype("category")
        commodity_data.Commodity_Group = commodity_data.Commodity_Group.astype("category")
        return commodity_data
    
    def read_historic_data(self):
        data = pd.read_csv(self.ADDINFOPATH / toolbox_paths.HISTINFO)
        data = self.downcasting(data)
        return data
        
    def downcasting(self, data: pd.DataFrame):
        data.RegionCode = data.RegionCode.astype("category")
        data.CommodityCode = data.CommodityCode.astype("category")
        data.domain = data.domain.astype("category")
        data.price = data.price.astype("float32")
        data.quantity = data.quantity.astype("float32")
        data.Period = data.Period.astype("int16")
        data.year = data.year.astype("int16")
        data.Scenario = data.Scenario.astype("category")
        data.Model = data.Model.astype("category")
        return data
    
    def add_consumption(self, data):
        data["quantity"] = (data["quantity_ManufactureCost"] +
                            data["quantity_Supply"] -
                            data["quantity_TransportationExport"] +
                            data["quantity_TransportationImport"])
        data.loc[data["quantity"] < 0, "quantity"] = 0
        data["price"] = (((data["quantity_ManufactureCost"] * data["price_ManufactureCost"]) +
                          (data["quantity_Supply"] * data["price_Supply"]) -
                          (data["quantity_TransportationExport"] * data["price_TransportationExport"]) +
                          (data["quantity_TransportationImport"]* data["price_TransportationImport"]))/
                          data["quantity"])
        data["price"] = 0
        data["domain"] = "Consumption"
        return data
    
    def add_net_exports(self, data):
        data["quantity"] =  (data["quantity_TransportationExport"] -
                            data["quantity_TransportationImport"])
        data["price"] =  data["price_TransportationExport"]
        data["domain"] = "Net Exports"
        return data
    
    def add_net_imports(self, data):
        data["quantity"] =  (data["quantity_TransportationImport"] -
                            data["quantity_TransportationExport"])
        data["price"] = data["price_TransportationImport"]
        data["domain"] = "Net Imports"
        return data
    
    def add_production(self, data):
        data["quantity"] =  (data["quantity_ManufactureCost"] + data["quantity_Supply"])
        data["price"] = (((data["quantity_ManufactureCost"] * data["price_ManufactureCost"]) +
                          (data["quantity_Supply"] * data["price_Supply"])) / data["quantity"])
        # if data["price"].mean() <=0:
        #     data["price"] = 0
        data["domain"] = "Production"
        return data
    
    def concat_calc_domains(self,origin_data:pd.DataFrame,calc_data:pd.DataFrame):
        calc_data = calc_data[['RegionCode','CommodityCode','Period','year','domain','price','quantity']].reset_index(drop=True) 
        result_df = pd.concat([origin_data, calc_data], axis=0).reset_index(drop=True) 
        return result_df

    def add_calculated_domains(self,data:pd.DataFrame):
        pivoted_price = data["data_periods"].pivot(index=["RegionCode", "CommodityCode", "Period", "year"], 
                         columns="domain", 
                         values="price").add_prefix("price_")
        pivoted_quantity = data["data_periods"].pivot(index=["RegionCode", "CommodityCode", "Period", "year"], 
                                    columns="domain", 
                                    values="quantity").add_prefix("quantity_")
        pivoted_df = pd.concat([pivoted_price, pivoted_quantity], axis=1).reset_index()    

        calculated_functions = [#"add_consumption",
                                "add_net_exports",
                                "add_net_imports",
                                #"add_production",
                                ]
        for method_name in calculated_functions:
            calc_df = getattr(self, method_name)(data=pivoted_df)
            data["data_periods"] = self.concat_calc_domains(origin_data=data["data_periods"], calc_data=calc_df)

        return data["data_periods"]

    def concat_scenarios(self, data: pd.DataFrame, sc_name:str, data_prev: pd.DataFrame, ID: int):
        """concat_scenarios, add scenario name from pkl file to data frames
        :param data: dictionary of the data container
        :param sc_name: scenario name from file name in dictionary
        """    
        data["data_periods"] = self.add_calculated_domains(data=data)
        try:
            for key in data: #loop through all data from datacontainer
                data[key][toolbox_parameters.column_name_scenario] = sc_name
                data[key][toolbox_parameters.column_name_model] = toolbox_parameters.model_name
                #data[key][parameters.column_name_id.value] = ID
                if data_prev != []:
                    data[key] = pd.concat([data_prev[key], data[key]], axis=0)
        except KeyError:
            pass
                
    def combined_data(self):
        """loop trough all input files in input directory
        """
        scenario_path = self.SCENARIOPATH
        num_files_to_read = self.num_files_to_read
        pkl_files = [
            Path(scenario_path) / file
            for file in os.listdir(scenario_path)
            if file.endswith(".pkl")
        ]
        sorted_files = sorted(pkl_files, key=lambda x: x.stat().st_mtime, reverse=True)
        newest_files = sorted_files[:num_files_to_read]

        data = []
        data_prev = []
        ID = 1
        for scenario_files in newest_files:
            src_filepath = scenario_path / scenario_files
            print(src_filepath)
            scenario_name = str(scenario_files)[str(scenario_files).rfind(toolbox_parameters.seperator_scenario_name)+3
                                        :-4]
            try:
                with gzip.open(src_filepath,'rb') as f:
                    if type(f) == gzip.GzipFile:
                        data = pickle.load(f)
                        data['data_periods'] = data['data_periods'][['RegionCode','CommodityCode','Period','year','domain','price','quantity']]
                self.concat_scenarios(data=data, sc_name=scenario_name, data_prev=data_prev, ID=ID)
            except gzip.BadGzipFile:
                pass
            except pickle.UnpicklingError:
                pass
            except PermissionError:
                pass
            except ValueError:
                pass

            data_prev = data
            ID += 1
        
        data_prev["data_periods"] = self.downcasting(data_prev["data_periods"])
        try:
            data = self.read_historic_data()
        except FileNotFoundError:
            data = pd.DataFrame()
        country_data = self.read_country_data()
        commodity_data = self.read_commodity_data()
        forest_data = data_prev['Forest']
        forest_data = forest_data[['Scenario','RegionCode','Period','ForStock','ForArea']]
        forest_data = forest_data.drop_duplicates(subset=['Scenario', 'RegionCode', 'Period'], keep='first')
        data_prev["data_periods"] = pd.merge(data_prev["data_periods"], forest_data, how='left', on=['Scenario','RegionCode','Period'])
        data_prev["data_periods"] = pd.concat([data_prev["data_periods"], data], axis=0)
        data_prev["data_periods"] = pd.merge(data_prev["data_periods"], country_data, on="RegionCode", how="left")
        data_prev["data_periods"] = pd.merge(data_prev["data_periods"], commodity_data, on="CommodityCode", how="left")
        data_prev["data_periods"]["domain"] = data_prev["data_periods"]["domain"].replace({
            'ManufactureCost': 'Manufacturing',
            'TransportationExport': 'Export',
            'TransportationImport': 'Import',
            })
        data_prev["data_periods"] = data_prev["data_periods"][['Model','Scenario','RegionCode','Continent','Country','ISO3',
                                                               'CommodityCode','Commodity','Commodity_Group','Period','year',
                                                               'domain','price','quantity',
                                                               'ForStock','ForArea',
                                                               ]]
        return data_prev

    def read_forest_data_gfpm(self, country_data:pd.DataFrame):
        for_data_gfpm = pd.read_csv(self.ADDINFOPATH / toolbox_paths.FORESTINFO, encoding = "ISO-8859-1")
        
        rearranged_for_data = pd.melt(for_data_gfpm, id_vars=['domain','Country'], var_name='Year',value_name='for')
        rearranged_for_data = rearranged_for_data.dropna()
        rearranged_for_data['Year'] = rearranged_for_data['Year'].astype(int)

        foreststock = pd.DataFrame()        
        for domain in rearranged_for_data.domain.unique():
            rearranged_for_data_domain = rearranged_for_data[rearranged_for_data['domain'] == domain].reset_index(drop=True)
            if domain == 'ForArea':
                rearranged_for_data_domain['ForStock'] = foreststock
            else: 
                foreststock = rearranged_for_data_domain['for']
        forest_data = rearranged_for_data_domain[['Country', 'Year', 'for', 'ForStock']]
        forest_data.columns = ['Country', 'Year', 'ForArea', 'ForStock']
        forest_data = pd.merge(forest_data, country_data, on= 'Country')

        period_mapping = {2017: 0, 2020: 1, 2025: 2, 2030: 3, 2035: 4, 2040: 5, 2045: 6, 2050: 7, 2055: 8, 2060: 9, 2065: 10}
        forest_data['Period'] = forest_data['Year'].map(period_mapping)

        forest_gfpm = forest_data[['RegionCode', 'Period', 'ForStock', 'ForArea']]
        forest_gfpm[toolbox_parameters.column_name_scenario]= 'world500'
        forest_data['Model'] = 'GFPM'
        return forest_gfpm
    
if __name__ == "__main__":
    import_pkl = import_pkl_data()
    data = import_pkl.combined_data()
    print(data['data_periods'])


import plotly.graph_objects as go
import dash
import dash_bootstrap_components as dbc
from dash import dcc, html
from dash.dependencies import Input, Output, State
import webbrowser
from threading import Timer
import pandas as pd
import numpy as np
from pathlib import Path
import plotly.express as px
import textwrap

PACKAGEDIR = Path(__file__).parent.parent.absolute()

class DashboardPlotter:

    def __init__(self, data):
        self.data = data
        self.app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        self.start = self.data['year'].min()
        self.end = self.data['year'].max()
        self.color_list = [
            '#2A4D69',  # Dunkelblau
            '#4B8BBE',  # Hellblau
            '#D35400',  # Dunkelorange
            '#AAB7B8',  # Grau
            '#9B59B6',  # Lila
            '#2980B9',  # Blau
            '#27AE60',  # Grün
            '#6C757D',  # Dunkelgrau
            '#F1C40F',  # Senfgelb
            '#E67E22',  # Orange  
        ]
        self.create_layout()
        self.create_callbacks()

    def create_layout(self):
        dropdown_style = {'height': '30px','marginBottom': '10px'}
        self.app.layout = dbc.Container([
            dbc.Row([
                dbc.Col(width=5),  # Leere Spalte für den linken Rand
                dbc.Col(
                    html.Img(src="https://raw.githubusercontent.com/TI-Forest-Sector-Modelling/TiMBA/ToolBox_implementation_cm/images/timba_dashboard_logo.png",
                             style={'height': '90px', 'width': 'auto'}),
                    width=1
                ),
                dbc.Col(
                    #html.H1("TiMBA Dashboard", className="text-center mb-4"), width=10
                )
            ]),
            dbc.Row([
                dbc.Col([
                    dbc.Card([
                        dbc.CardBody([
                            html.H4("Filters", className="card-title"),
                            dcc.Dropdown(id='region-dropdown',
                                         options=[{'label': i, 'value': i} for i in sorted(self.data['ISO3'].dropna().unique())],
                                         multi=True,
                                         placeholder="Select Country...",
                                         style=dropdown_style),
                            dcc.Dropdown(id='continent-dropdown',
                                         options=[{'label': i, 'value': i} for i in sorted(self.data['Continent'].dropna().unique())],
                                         placeholder="Select Continent...",
                                         multi=True,
                                         style=dropdown_style),
                            dcc.Dropdown(id='domain-dropdown',
                                         options=[{'label': i, 'value': i} for i in sorted(self.data['domain'].dropna().unique())],
                                         placeholder="Select Domain...",
                                         multi=True,
                                         style=dropdown_style),
                            dcc.Dropdown(id='commodity-dropdown',
                                         options=[{'label': i, 'value': i} for i in sorted(self.data['Commodity'].dropna().unique())],
                                         placeholder="Select Commodity...",
                                         multi=True,
                                         style=dropdown_style),
                            dcc.Dropdown(id='commodity-group-dropdown',
                                         options=[{'label': i, 'value': i} for i in self.data['Commodity_Group'].dropna().unique().tolist()],
                                         placeholder="Select Commodity Group...",
                                         multi=True,
                                         style=dropdown_style),
                            dcc.Dropdown(id='scenario-filter',
                                         options=[{'label': i, 'value': i} for i in self.data['Scenario'].unique()],
                                         placeholder="Select Scenario...",
                                         multi=True,
                                         style=dropdown_style),
                            html.Button("Download CSV", id="btn_csv"),
                            dcc.Download(id="download-dataframe-csv"),
                        ])
                    ], className="mb-4", style={'white': 'white'}),  #filter box
                    dbc.Card([
                        dbc.CardBody([
                                dcc.Graph(id='price-plot',
                                      config={'toImageButtonOptions': {'format': 'png', 'filename': 'price_plot'}},
                                      style={'height': '45vh','width':'46vh'})
                    ])
                ], style={'white': 'white'})  #price box
            ], width=3),
            dbc.Col([
                dbc.Row([
                    dbc.Col([
                        dbc.Card([
                            dbc.CardBody([
                                dcc.Graph(id='quantity-plot',
                                          config={'toImageButtonOptions': {'format': 'png', 'filename': 'quantity_plot'}},
                                          style={'height': '84.5vh'})
                            ])
                        ], style={'backgroundColor': 'white'}) #mittelbox
                    ], width=8), # Breite auf 8 reduziert
                    dbc.Col([
                        dbc.Card([
                            dbc.CardBody([
                                dcc.Graph(id='forstock-plot',  # Geändert: ID auf 'forstock-plot'
                                          config={'toImageButtonOptions': {'format': 'png'}},
                                          style={'height': '39vh'})
                            ])
                        ], style={'backgroundColor': 'white', 'marginBottom': '20px'}), # Abstand hinzugefügt
                        dbc.Card([
                            dbc.CardBody([
                                html.H5("Filter for Worldmap", className="card-title"),
                                #html.H6("Scenario Filter", className="card-title"),  # Titel für den Scenario-Filter
                                #html.H6("Year Filter", className="card-title"),  # Titel für den Year-Filter
                                dcc.Dropdown(
                                    id='year-filter',
                                    options=[{'label': i, 'value': i} for i in sorted(self.data['year'].unique())],
                                    placeholder="Select Year...",
                                    style=dropdown_style
                                ),
                                 dcc.Graph(id='world-map',  # Geändert: ID auf 'world-map'
                                          config={'toImageButtonOptions': {'format': 'png'}},
                                          style={'height': '31.75vh'})
                            ])
                        ], style={'backgroundColor': 'white'})
                    ], width=4), # Breite auf 4 gesetzt
                ]),
            ], width=9)
        ])
    ], fluid=True, style={'backgroundColor': 'white'}) #gesamthintergrund

    def create_callbacks(self):
        @self.app.callback(
            [Output('quantity-plot', 'figure'),
             Output('price-plot', 'figure'),
             Output('forstock-plot', 'figure')],
            [Input('region-dropdown', 'value'),
             Input('continent-dropdown', 'value'),
             Input('domain-dropdown', 'value'),
             Input('commodity-dropdown', 'value'),
             Input('commodity-group-dropdown', 'value'),
             Input('scenario-filter', 'value')]
        )
        def update_plots(region, continent, domain, commodity, commodity_group,scenario):
            return self.update_plot_data(region, continent, domain, commodity, commodity_group,scenario)

        @self.app.callback(
            Output('world-map', 'figure'),
            [Input('scenario-filter', 'value'),
             Input('year-filter', 'value'),
             Input('region-dropdown', 'value'),
             Input('continent-dropdown', 'value'),
             Input('domain-dropdown', 'value'),
             Input('commodity-dropdown', 'value'),
             Input('commodity-group-dropdown', 'value')]
        )
        def update_world_map(scenario, year, region, continent, domain, commodity, commodity_group):
            return self.create_world_map(region, continent, domain, commodity, commodity_group, scenario, year)

        @self.app.callback(
            Output("download-dataframe-csv", "data"),
            Input("btn_csv", "n_clicks"),
            [State('region-dropdown', 'value'),
            State('continent-dropdown', 'value'),
            State('domain-dropdown', 'value'),
            State('commodity-dropdown', 'value'),
            State('commodity-group-dropdown', 'value'),
            State('scenario-filter', 'value')],  # Geändert von Input zu State
            prevent_initial_call=True
        )
        def func(n_clicks, region, continent, domain, commodity, commodity_group, scenario):
            if n_clicks is None:
                raise dash.exceptions.PreventUpdate
            filtered_data = self.filter_data(region, continent, domain, commodity, commodity_group, scenario)
            return dcc.send_data_frame(filtered_data.to_csv, "filtered_data.csv")


    def filter_data(self, region, continent, domain, commodity, commodity_group, scenario):
        filtered_data = self.data
        if region and isinstance(region, list):
            filtered_data = filtered_data[filtered_data['ISO3'].isin(region)]
        if continent and isinstance(continent, list):
            filtered_data = filtered_data[filtered_data['Continent'].isin(continent)]
        if domain and isinstance(domain, list):
            filtered_data = filtered_data[filtered_data['domain'].isin(domain)]
        if commodity and isinstance(commodity, list):
            filtered_data = filtered_data[filtered_data['Commodity'].isin(commodity)]
        if commodity_group and isinstance(commodity_group, list):
            filtered_data = filtered_data[filtered_data['Commodity_Group'].isin(commodity_group)]
        if scenario and isinstance(scenario, list):
            filtered_data = filtered_data[filtered_data['Scenario'].isin(scenario)]
        filtered_data = self.remove_extreme_outliers(df=filtered_data, col='price')
        return filtered_data

    def update_plot_data(self, region, continent, domain, commodity, commodity_group, scenario):
        graphic_template='plotly_white'#'plotly_dark'#'plotly_white'
        filtered_data = self.filter_data(region, continent, domain, commodity, commodity_group, scenario)

        # Quantity plot
        grouped_data_quantity = filtered_data.groupby(['year', 'Scenario']).sum().reset_index()
        grouped_data_quantity = grouped_data_quantity[(grouped_data_quantity["year"] >= self.start) & (grouped_data_quantity["year"] <= self.end)]
        fig_quantity = go.Figure()
        for i, scenario in enumerate(grouped_data_quantity['Scenario'].unique()):
            subset = grouped_data_quantity[grouped_data_quantity['Scenario'] == scenario]
            color = self.color_list[i % len(self.color_list)]
            dash = 'solid' if scenario in ['Historic Data'] else 'dash'
            fig_quantity.add_trace(go.Scatter(x=subset['year'], y=subset['quantity'], mode='lines',
                                              name=f'{scenario}', line=dict(color=color, dash=dash)))
        title_quantity = self.generate_title(region, continent, domain, commodity, commodity_group)
        title= "Quantity by Period and Scenario for " + title_quantity
        fig_quantity.update_layout(
            title='<br>'.join(textwrap.wrap(title, width=90)),
            xaxis_title='Year',
            yaxis_title='Quantity',
            yaxis=dict(rangemode='nonnegative', zeroline=True, zerolinewidth=2, zerolinecolor='LightGrey'),
            legend_title='Scenario',
            legend=dict(orientation="h", yanchor="top", y=-0.1, xanchor="center", x=0.5),
            margin=dict(l=35, r=35, t=60, b=90),
            hovermode='x unified',
            template=graphic_template
        )

        # Price plot
        grouped_data_price = filtered_data.groupby(['year', 'Scenario']).mean().reset_index()
        max_year = grouped_data_price['year'].max() + 0.5

        fig_price = go.Figure()
        for i, scenario in enumerate(grouped_data_price['Scenario'].unique()):
            subset = grouped_data_price[grouped_data_price['Scenario'] == scenario]
            color = self.color_list[i % len(self.color_list)]
            fig_price.add_trace(go.Bar(x=subset['price'], y=subset['year'], orientation='h',
                                       name=f'{scenario}', marker_color=color))
            
        title_price = f'Price by Period and Scenario'
        fig_price.update_layout(
            title=title_price,
            xaxis_title='Price',
            yaxis_title='Year',
            yaxis=dict(range=[2020-0.5, max_year]),
            legend_title='Scenario',
            template=graphic_template,
            showlegend=False,
            #legend=dict(orientation="h", yanchor="top", y=-0.25, xanchor="center", x=0.5),
            margin=dict(l=35, r=60, t=50, b=5),
            barmode='group'
        )

        # ForStock plot
        grouped_data_stock = filtered_data.drop(columns=['domain', 'price','quantity','CommodityCode','Commodity','Commodity_Group'])
        grouped_data_stock = grouped_data_stock.drop_duplicates().reset_index(drop=True)
        grouped_data_stock = grouped_data_stock.groupby(['year', 'Scenario']).agg({
            'ForStock': 'sum',
            }).reset_index()
        grouped_data_stock = grouped_data_stock[grouped_data_stock.Scenario!='Historic Data']
        fig_stock = go.Figure()
        for i, scenario in enumerate(grouped_data_stock['Scenario'].unique()):
            subset = grouped_data_stock[grouped_data_stock['Scenario'] == scenario]
            color = self.color_list[i+1 % len(self.color_list)]
            fig_stock.add_trace(go.Bar(x=subset['year'], y=subset['ForStock'],
                                    name=f'{scenario}', marker_color=color))
            
        min_year = 2020 - 1
        max_year = grouped_data_stock['year'].max() + 0.5

        min_val = grouped_data_stock['ForStock'].min() * 0.9
        max_val = grouped_data_stock['ForStock'].max() * 1.1

        fig_stock.update_layout(
            title='Forest Stock by Year and Scenario',
            xaxis_title='Year',
            xaxis=dict(range=[min_year, max_year]),
            yaxis=dict(range=[min_val, max_val]),
            yaxis_title='ForStock',
            template=graphic_template,
            showlegend=False,
            #legend=dict(orientation="h", yanchor="top", y=-0.35, xanchor="center", x=0.5),
            margin=dict(l=50, r=50, t=40, b=5),
            barmode='group'
        )

        return fig_quantity, fig_price, fig_stock

    def generate_title(self, region, continent, domain, commodity, commodity_group):
        title_parts = []
        if region:
            title_parts.append(f"{region}")
        if continent:
            title_parts.append(f"{continent}")
        if domain:
            title_parts.append(f"{domain}")
        if commodity:
            title_parts.append(f"{commodity}")
        if commodity_group:
            title_parts.append(f"{commodity_group}")
        title = ", ".join(title_parts) if title_parts else "all data"
        clean_title = title.replace("'", "").replace("[", "").replace("]", "")
        return clean_title

    def create_world_map(self, region, continent, domain, commodity, commodity_group, scenario=None, year=None):
        filtered_data = self.filter_data(region, continent, domain, commodity, commodity_group, scenario)
        if year:
            filtered_data = filtered_data[filtered_data['year']==year]
        country_data = filtered_data.groupby('ISO3')['quantity'].sum().reset_index()
        country_data = country_data[country_data['quantity']>=0.001].reset_index()

        fig = px.choropleth(
            country_data,
            locations="ISO3",
            color="quantity",
            hover_name="ISO3",
            color_continuous_scale="Greens"
        )

        title = 'Worldmap for '+ self.generate_title(region, continent, domain, commodity, commodity_group)
        fig.update_layout(
            #title= '<br>'.join(textwrap.wrap(title, width=43)),
            geo=dict(
                showcoastlines=True,
                coastlinecolor="LightGray",
                showocean=False,
                oceancolor="LightBlue",
                projection_type='natural earth',
                # Grenzen bestimmen
                lonaxis_range=[-360, 360],  # Längengradbereich
                lataxis_range=[-55, 55],   # Breitengradbereich
            ),
            margin=dict(l=1, r=1, t=1, b=1),  # Ränder minimieren
            coloraxis_showscale=False  # Legende entfernen
        )

        return fig
    
    def get_last_historic_year(self):
        historic_data = self.data[self.data['Scenario'] == 'Historic Data']
        if not historic_data.empty:
            return historic_data['year'].max()
        else:
            return self.data['year'].max()
    
    def remove_extreme_outliers(self,df:pd.DataFrame,col:str,threshhold:float=50):        
        Q1 = df[col].quantile(0.25)
        Q3 = df[col].quantile(0.75)
        IQR = Q3 - Q1
        outlier_threshold = threshhold * IQR
        df.loc[df[col] >= outlier_threshold, col] = np.nan
        return df
    
    def open_browser(self):
        webbrowser.open_new("http://localhost:8050")


    def run(self):
        Timer(1, self.open_browser).start()
        self.app.run(host='localhost', debug=False, dev_tools_ui=False, dev_tools_hot_reload=False, port=8050)



from Toolbox.classes.import_data import import_pkl_data
from Toolbox.classes.dashboard import DashboardPlotter
from pathlib import Path
import Toolbox.parameters.paths as toolbox_paths

class timba_dashboard:
    def __init__(self,scenario_folder_path:Path,
                 additional_info_folderpath:Path,
                 num_files_to_read:int=10):
        self.num_files_to_read = num_files_to_read
        self.scenario_folder_path = scenario_folder_path
        self.additional_info_folderpath = additional_info_folderpath

    def import_data(self):
        import warnings
        warnings.simplefilter(action='ignore', category=FutureWarning)
        import_pkl = import_pkl_data(num_files_to_read=self.num_files_to_read,
                                     SCENARIOPATH=self.scenario_folder_path,
                                     ADDINFOPATH=self.additional_info_folderpath)
        self.data = import_pkl.combined_data()

    def call_dashboard(self):
        DashboardPlotter(data=self.data["data_periods"]).run()

    def run(self):
        self.import_data()
        self.call_dashboard() 

if __name__ == "__main__":
    td = timba_dashboard(num_files_to_read=4,
                         scenario_folder_path=toolbox_paths.SCINPUTPATH)
    td.run()

